// Copyright 2016 CERN and GSI
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
#ifdef MODBUS_SUPPORT_ENABLED
#include <silecs-communication/interface/utility/SilecsLog.h>
#include <silecs-communication/interface/utility/StringUtilities.h>
#include <silecs-communication/interface/communication/SilecsConnection.h>
#include <silecs-communication/interface/equipment/SilecsPLC.h>
#include <silecs-communication/interface/communication/MBConnection.h>
#include <silecs-communication/interface/utility/SilecsException.h>

//Modbus max data size definition (byte counting)
#define MAX_WRITE_DATA_SIZE MODBUS_MAX_WRITE_REGISTERS*2
#define MAX_READ_DATA_SIZE  MODBUS_MAX_READ_REGISTERS*2

namespace Silecs
{

MBConnection::MBConnection(PLC* thePLC) :
                Connection(thePLC)
{
    LOG(ALLOC) << "MBConnection (create): " << thePLC->getName();

    //Connection use IP address to limit the naming-server accesses
    readCtx_ = modbus_new_tcp((char *)thePLC->getIPAddress().c_str(), MODBUS_TCP_DEFAULT_PORT /*=502*/);
    writeCtx_ = modbus_new_tcp((char *)thePLC->getIPAddress().c_str(), MODBUS_TCP_DEFAULT_PORT /*=502*/);

    /*TODO: To be adjusted with the next stable libmodbus release (>3.1.1)
     which will fix the current timeout response time issue (see libmodbus forum).

     Define Modbus response timeout
     struct timeval response_timeout;
     response_timeout.tv_sec = 0;
     response_timeout.tv_usec = 10000;

     modbus_set_response_timeout(readCtx_ , &response_timeout);
     modbus_set_response_timeout(writeCtx_ , &response_timeout);
     */

    modbus_set_slave(readCtx_, 1);
    //modbus_set_debug(readCtx_, TRUE);
}

MBConnection::~MBConnection()
{
    //Close the connection before removing resources
    //disable(); must be done before removing resource
    modbus_free(writeCtx_);
    modbus_free(readCtx_);
}

bool MBConnection::open(PLC* thePLC)
{
    int readErr = modbus_connect(readCtx_);
    int writeErr = modbus_connect(writeCtx_);
    return ( (readErr != -1) && (writeErr != -1));
}

bool MBConnection::close(PLC* thePLC)
{
    modbus_close(readCtx_);
    modbus_close(writeCtx_);
    return true;
}

/*----------------------------------------------------------*/
int MBConnection::mbWriteFrames(modbus_t* ctx, long start_addr, uint16_t count, uint8_t* data)
{
    uint16_t *srcp = (uint16_t*)data;
    int word_count, byte_count;

    while (count)
    {
        word_count = (count > MAX_WRITE_DATA_SIZE) ? MAX_WRITE_DATA_SIZE / 2 : (count / 2) + (count % 2);
        byte_count = (word_count * 2);

        if (modbus_write_registers(ctx, start_addr, word_count, srcp) != word_count)
        {
            return -1;
        }

        //srcp += byte_count-(count%2);
        srcp += word_count;
        start_addr += word_count;
        count -= (byte_count - (count % 2));
    }

    return 0;
}

/*----------------------------------------------------------*/
int MBConnection::mbReadFrames(modbus_t* ctx, long start_addr, uint16_t count, uint8_t* data)
{
    uint16_t *destp = (uint16_t*)data;
    int word_count, byte_count;

    while (count)
    {

        if (count > MAX_READ_DATA_SIZE)
        {
            /*'word_count' is a word counter*/
            word_count = MAX_READ_DATA_SIZE / 2;
        }
        else
        {
            /*'word_count' is a word counter*/
            word_count = (count / 2) + (count % 2);
        }

        byte_count = (word_count * 2);

        if (modbus_read_registers(ctx, start_addr, word_count, destp) != word_count)
        {
            return -1;
        }

        //destp += (byte_count-(count%2));
        destp += word_count;
        start_addr += word_count;
        count -= (byte_count - (count % 2));
    }

    return 0;
}

int MBConnection::readData(PLC* thePLC, long address, unsigned long offset, unsigned long size, unsigned char* pBuffer)
{
    int err = 0;

    //Schneider uses 16bit alignment memory. Block address is expressed in bytes, must be an even value!
    if (address % 2)
        throw SilecsException(__FILE__, __LINE__, PARAM_INCORRECT_BLOCK_ADDRESS, StringUtilities::toString(address));

    /* . There is one read-channel per PLC connection. It must be protected against concurrent access.
     * . The write-channel is independent and can be accessed in parallel.
     * . The global action (open,close,etc.) must be protected from any concurrent access.
     * Attention!
     * Mutexes are defined with recursive option. It allows doOpen method re-calling readData
     * method by executing register synchronization if required.
     */

    //(re)connect the PLC if needed and (re)synchronize the retentive registers
    if (doOpen(thePLC))
    {
        //connection is established then acquire data
        readLock();
        //Schneider uses 16bit alignment memory. Block address is expressed in bytes (==> /2)
        unsigned short addr = ((unsigned short)address + (unsigned short)offset) / 2;

        //DATA topic makes sense with RECV one
        if (RECV & Log::topics_)
            LOG(DATA) << "Read data, address: %MW" << addr << ", byte-size: " << size;

        err = mbReadFrames(readCtx_, addr, (unsigned short)size, pBuffer);
        checkError(thePLC, err, false); // close the connection, will try again at the next access
        readUnlock();
    }
    return err;
}

int MBConnection::writeData(PLC* thePLC, long address, unsigned long offset, unsigned long size, unsigned char* pBuffer)
{
    int err = 0;

    //Schneider uses 16bit alignment memory. Block address is expressed in bytes, must be an even value!
    if (address % 2)
        throw SilecsException(__FILE__, __LINE__, PARAM_INCORRECT_BLOCK_ADDRESS, StringUtilities::toString(address));

    /* . There is one read-channel per PLC connection. It must be protected against concurrent access.
     * . The write-channel is independent and can be accessed in parallel.
     * . The global action (open,close,etc.) must be protected from any concurrent access.
     * Attention!
     * Mutexes are defined with recursive option. It allows doOpen method re-calling sendData
     * method by executing register synchronization if required.
     */

    //(re)connect the PLC if needed and (re)synchronize the retentive registers
    if (doOpen(thePLC))
    {
        //connection is established then send data
        writeLock();
        //Schneider uses 16bit alignment memory. Block address is expressed in bytes (==> /2)
        unsigned short addr = ((unsigned short)address + (unsigned short)offset) / 2;

        //DATA topic makes sense with SEND one
        if (SEND & Log::topics_)
            LOG(DATA) << "Write data, address: %MW" << addr << ", byte-size: " << size;

        err = mbWriteFrames(writeCtx_, (unsigned short)addr, (unsigned short)size, pBuffer);
        checkError(thePLC, err, false); // close the connection, will try again at the next access
        writeUnlock();
    }
    return err;
}

int MBConnection::readMemory(PLC* thePLC, long address, unsigned long offset, unsigned long size, unsigned char* pBuffer)
{
    return readData(thePLC, address, offset, size, pBuffer);
}

int MBConnection::writeMemory(PLC* thePLC, long address, unsigned long offset, unsigned long size, unsigned char* pBuffer)
{
    return writeData(thePLC, address, offset, size, pBuffer);
}

int MBConnection::readIO(PLC* thePLC, long address, unsigned long offset, unsigned long size, unsigned char* pBuffer)
{
    return readData(thePLC, address, offset, size, pBuffer);
}

int MBConnection::writeIO(PLC* thePLC, long address, unsigned long offset, unsigned long size, unsigned char* pBuffer)
{
    return writeData(thePLC, address, offset, size, pBuffer);
}

//-------------------------------------------------------------------------------------------------------------------
bool MBConnection::checkError(PLC* thePLC, int err, bool retry)
{

    if (err != 0)
    {
        LOG(COMM) << "Transaction failure with the controller: " << thePLC->getName() << ". MODBUS[" << errno << "]: " << modbus_strerror(errno);

        if (retry)
        {
            LOG(COMM) << "Try to reconnect the controller: " << thePLC->getName();
            if (reOpen(thePLC))
            { // can repeat the request after the connection was successfully reopened
                return true;
            }
            // reconnection has failed again, just close the connection
            LOG(COMM) << "Unable to reconnect the controller: " << thePLC->getName();
        }
        // no retry, we just want to close (use default case)
        doClose(thePLC, /*withLock =*/true);
    }
    return false;
}

} // namespace

#endif //MODBUS_SUPPORT_ENABLED
